"""Sample file for defining new units.

Note that this file will be overwritten when you upgrade to a new version
of Unum, so if you have custom units they should be in their own file.
"""

from robotpy_toolkit_7407.unum import Unum
unit = Unum.unit

# << uncomment this import if you want to derive your units from SI >>
# from robotpy_toolkit_7407.unum.units.si import *

# << define your units hereafter, e.g.
#    M  = unit(  'm' , 0          , 'meter'     )
#    KM = unit( 'km' , 1000. * M  , 'kilometer' ) >>




# cleaning
del Unum
del unit



import datetime
import inspect
import os

import robotpy_toolkit_7407.utils.logger as lg


class Logger:
    def __init__(
        self,
        debug: bool = False,
        filename: str = f"custom_logs/custom_logging_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log",
    ):
        self.filename: str = filename
        self.logfile = None
        self.debug_on = debug

        self.start_time = datetime.datetime.now()

    def log(self, system: str, message: str):
        self.logfile = open(self.filename, "a")

        frame = inspect.stack()[1].frame
        file_name = os.path.basename(frame.f_code.co_filename)
        line_no = str(frame.f_lineno)

        self.logfile.write(
            f"[{str(datetime.datetime.now() - self.start_time) + ']'} [{file_name + ':' + line_no + ']' : <19} [{system + ']'  : <15} ~ {message  : <20}\n"
        )
        self.logfile.close()

        lg.info(message, system, frame)

    def debug(self, system: str, message: str):
        if self.debug_on:
            self.logfile = open(self.filename, "a")

            frame = inspect.stack()[1].frame
            file_name = os.path.basename(frame.f_code.co_filename)
            line_no = str(frame.f_lineno)

            self.logfile.write(
                f"[{str(datetime.datetime.now() - self.start_time) + ']'} [{file_name + ':' + line_no + ']' : <19} [{system + ']'  : <15} ~ {message  : <20}\n"
            )
            self.logfile.close()

            lg.info(message, system, frame)

    def close(self):
        self.logfile.close()



import wpilib


class Compressor:
    """
    Wrapper class for a wpilib compressor.
    """

    def __init__(self, module, ctre_hub=False):
        """
        Args:
            module (int): The Pneumatics Control Module (PCM) CAN ID
            ctre_hub (bool): Whether the PCM is connected to a CTRE Pneumatics Control Module or not
        """
        module_type = wpilib.PneumaticsModuleType.CTREPCM if ctre_hub else wpilib.PneumaticsModuleType.REVPH
        self.compressor = wpilib.Compressor(module, module_type)

    def get_compressor_enabled(self) -> bool:
        """
        Get whether the compressor is enabled.
        """
        return self.compressor.enabled()

    def get_current(self) -> float:
        """
        Get the current draw of the compressor.
        """
        return self.compressor.getCurrent()

    def enable_analog(self, minimum_psi: float, maximum_psi: float):
        """
        Enable the compressor.

        Args:
            minimum_psi (float): The minimum pressure in PSI to enable the compressor
            maximum_psi (float): The maximum pressure in PSI to disable the compressor
        """
        self.compressor.enableAnalog(minimum_psi, maximum_psi)

    def start(self):
        """
        Start the compressor.
        """
        self.compressor.start()

    def stop(self):
        """
        Stop the compressor.
        """
        self.compressor.stop()



import commands2


class Subsystem(commands2.SubsystemBase):
    """
    Extendable subsystem class. Needs to be extended by every subsystem.
    """

    def init(self):
        """
        Overridable method for initializing the subsystem. Place motor inits/other physical initialization here,
        for example re-zeroing a gyro.
        """
        ...



from robotpy_toolkit_7407.sensors.photonvision.photon_camera import PhotonCamera
from robotpy_toolkit_7407.sensors.photonvision.photon_odometry import PhotonOdometry, LoadFieldLayout
from robotpy_toolkit_7407.sensors.photonvision.photon_target import PhotonTarget, AprilTag


"""Define the seven SI base units.

Source : http://physics.nist.gov/cuu/Units/units.html)
"""

from robotpy_toolkit_7407.unum import Unum
unit = Unum.unit

m = M = unit("m", 0, "meter")
Ym = YM = unit("Ym", 10**24 * m, "yottameter")
Zm = ZM = unit("Zm", 10**21 * m, "zettameter")
Em = EM = unit("Em", 10**18 * m, "exameter")
Pm = PM = unit("Pm", 10**15 * m, "petameter")
Tm = TM = unit("Tm", 10**12 * m, "terameter")
Gm = GM = unit("Gm", 10**9 * m, "gigameter")
Mm = MM = unit("Mm", 10**6 * m, "megameter")
km = KM = unit("km", 10**3 * m, "kilometer")
hm = HM = unit("hm", 10**2 * m, "hectometer")
dam = DAM = unit("dam", 10**1 * m, "decameter")
ym = YM = unit("ym", 10**-24 * m, "yoctometer")
zm = ZM = unit("zm", 10**-21 * m, "zeptometer")
am = AM = unit("am", 10**-18 * m, "attometer")
fm = FM = unit("fm", 10**-15 * m, "femtometer")
pm = PM = unit("pm", 10**-12 * m, "picometer")
nm = NM = unit("nm", 10**-9 * m, "nanometer")
um = UM = unit("um", 10**-6 * m, "micrometer")
mm = MM = unit("mm", 10**-3 * m, "millimeter")
cm = CM = unit("cm", 10**-2 * m, "centimeter")
dm = DM = unit("dm", 10**-1 * m, "decimeter")

# Uppercase S is Siements; seconds can only use lowercase s
s = unit("s", 0, "second")
Ys = unit("Ys", 10**24 * s, "yottasecond")
Zs = unit("Zs", 10**21 * s, "zettasecond")
Es = unit("Es", 10**18 * s, "exasecond")
Ps = unit("Ps", 10**15 * s, "petasecond")
Ts = unit("Ts", 10**12 * s, "terasecond")
Gs = unit("Gs", 10**9 * s, "gigasecond")
Ms = unit("Ms", 10**6 * s, "megasecond")
ks = unit("ks", 10**3 * s, "kilosecond")
hs = unit("hs", 10**2 * s, "hectosecond")
das = unit("das", 10**1 * s, "decasecond")
ys = unit("ys", 10**-24 * s, "yoctosecond")
zs = unit("zs", 10**-21 * s, "zeptosecond")
#as = unit("as", 10**-18 * s, "attosecond") # as is a reserved word
fs = unit("fs", 10**-15 * s, "femtosecond")
ps = unit("ps", 10**-12 * s, "picosecond")
ns = unit("ns", 10**-9 * s, "nanosecond")
us = unit("us", 10**-6 * s, "microsecond")
ms = unit("ms", 10**-3 * s, "millisecond")
cs = unit("cs", 10**-2 * s, "centisecond")
ds = unit("ds", 10**-1 * s, "decisecond")


A = A = unit("A", 0, "ampere")
YA = YA = unit("YA", 10**24 * A, "yottaampere")
ZA = ZA = unit("ZA", 10**21 * A, "zettaampere")
EA = EA = unit("EA", 10**18 * A, "exaampere")
PA = PA = unit("PA", 10**15 * A, "petaampere")
TA = TA = unit("TA", 10**12 * A, "teraampere")
GA = GA = unit("GA", 10**9 * A, "gigaampere")
MA = MA = unit("MA", 10**6 * A, "megaampere")
kA = KA = unit("kA", 10**3 * A, "kiloampere")
hA = HA = unit("hA", 10**2 * A, "hectoampere")
daA = DAA = unit("daA", 10**1 * A, "decaampere")
yA = YA = unit("yA", 10**-24 * A, "yoctoampere")
zA = ZA = unit("zA", 10**-21 * A, "zeptoampere")
aA = AA = unit("aA", 10**-18 * A, "attoampere")
fA = FA = unit("fA", 10**-15 * A, "femtoampere")
pA = PA = unit("pA", 10**-12 * A, "picoampere")
nA = NA = unit("nA", 10**-9 * A, "nanoampere")
uA = UA = unit("uA", 10**-6 * A, "microampere")
mA = MA = unit("mA", 10**-3 * A, "milliampere")
cA = CA = unit("cA", 10**-2 * A, "centiampere")
dA = DA = unit("dA", 10**-1 * A, "deciampere")


K = K = unit("K", 0, "kelvin")
YK = YK = unit("YK", 10**24 * K, "yottakelvin")
ZK = ZK = unit("ZK", 10**21 * K, "zettakelvin")
EK = EK = unit("EK", 10**18 * K, "exakelvin")
PK = PK = unit("PK", 10**15 * K, "petakelvin")
TK = TK = unit("TK", 10**12 * K, "terakelvin")
GK = GK = unit("GK", 10**9 * K, "gigakelvin")
MK = MK = unit("MK", 10**6 * K, "megakelvin")
kK = KK = unit("kK", 10**3 * K, "kilokelvin")
hK = HK = unit("hK", 10**2 * K, "hectokelvin")
daK = DAK = unit("daK", 10**1 * K, "decakelvin")
yK = YK = unit("yK", 10**-24 * K, "yoctokelvin")
zK = ZK = unit("zK", 10**-21 * K, "zeptokelvin")
aK = AK = unit("aK", 10**-18 * K, "attokelvin")
fK = FK = unit("fK", 10**-15 * K, "femtokelvin")
pK = PK = unit("pK", 10**-12 * K, "picokelvin")
nK = NK = unit("nK", 10**-9 * K, "nanokelvin")
uK = UK = unit("uK", 10**-6 * K, "microkelvin")
mK = MK = unit("mK", 10**-3 * K, "millikelvin")
cK = CK = unit("cK", 10**-2 * K, "centikelvin")
dK = DK = unit("dK", 10**-1 * K, "decikelvin")


mol = MOL = unit("mol", 0, "mole")
Ymol = YMOL = unit("Ymol", 10**24 * mol, "yottamole")
Zmol = ZMOL = unit("Zmol", 10**21 * mol, "zettamole")
Emol = EMOL = unit("Emol", 10**18 * mol, "examole")
Pmol = PMOL = unit("Pmol", 10**15 * mol, "petamole")
Tmol = TMOL = unit("Tmol", 10**12 * mol, "teramole")
Gmol = GMOL = unit("Gmol", 10**9 * mol, "gigamole")
Mmol = MMOL = unit("Mmol", 10**6 * mol, "megamole")
kmol = KMOL = unit("kmol", 10**3 * mol, "kilomole")
hmol = HMOL = unit("hmol", 10**2 * mol, "hectomole")
damol = DAMOL = unit("damol", 10**1 * mol, "decamole")
ymol = YMOL = unit("ymol", 10**-24 * mol, "yoctomole")
zmol = ZMOL = unit("zmol", 10**-21 * mol, "zeptomole")
amol = AMOL = unit("amol", 10**-18 * mol, "attomole")
fmol = FMOL = unit("fmol", 10**-15 * mol, "femtomole")
pmol = PMOL = unit("pmol", 10**-12 * mol, "picomole")
nmol = NMOL = unit("nmol", 10**-9 * mol, "nanomole")
umol = UMOL = unit("umol", 10**-6 * mol, "micromole")
mmol = MMOL = unit("mmol", 10**-3 * mol, "millimole")
cmol = CMOL = unit("cmol", 10**-2 * mol, "centimole")
dmol = DMOL = unit("dmol", 10**-1 * mol, "decimole")


cd = CD = unit("cd", 0, "candela")
Ycd = YCD = unit("Ycd", 10**24 * cd, "yottacandela")
Zcd = ZCD = unit("Zcd", 10**21 * cd, "zettacandela")
Ecd = ECD = unit("Ecd", 10**18 * cd, "exacandela")
Pcd = PCD = unit("Pcd", 10**15 * cd, "petacandela")
Tcd = TCD = unit("Tcd", 10**12 * cd, "teracandela")
Gcd = GCD = unit("Gcd", 10**9 * cd, "gigacandela")
Mcd = MCD = unit("Mcd", 10**6 * cd, "megacandela")
kcd = KCD = unit("kcd", 10**3 * cd, "kilocandela")
hcd = HCD = unit("hcd", 10**2 * cd, "hectocandela")
dacd = DACD = unit("dacd", 10**1 * cd, "decacandela")
ycd = YCD = unit("ycd", 10**-24 * cd, "yoctocandela")
zcd = ZCD = unit("zcd", 10**-21 * cd, "zeptocandela")
acd = ACD = unit("acd", 10**-18 * cd, "attocandela")
fcd = FCD = unit("fcd", 10**-15 * cd, "femtocandela")
pcd = PCD = unit("pcd", 10**-12 * cd, "picocandela")
ncd = NCD = unit("ncd", 10**-9 * cd, "nanocandela")
ucd = UCD = unit("ucd", 10**-6 * cd, "microcandela")
mcd = MCD = unit("mcd", 10**-3 * cd, "millicandela")
ccd = CCD = unit("ccd", 10**-2 * cd, "centicandela")
dcd = DCD = unit("dcd", 10**-1 * cd, "decicandela")


kg = KG = unit("kg", 0, "kilogram")
Yg = YG = unit("Yg", 10**21 * kg, "yottagram")
Zg = ZG = unit("Zg", 10**18 * kg, "zettagram")
Eg = EG = unit("Eg", 10**15 * kg, "exagram")
Pg = PG = unit("Pg", 10**12 * kg, "petagram")
Tg = TG = unit("Tg", 10**9 * kg, "teragram")
Gg = GG = unit("Gg", 10**6 * kg, "gigagram")
Mg = MG = unit("Mg", 10**3 * kg, "megagram")
hg = HG = unit("hg", 10**-1 * kg, "hectogram")
dag = DAG = unit("dag", 10**-2 * kg, "decagram")
yg = YG = unit("yg", 10**-27 * kg, "yoctogram")
zg = ZG = unit("zg", 10**-24 * kg, "zeptogram")
ag = AG = unit("ag", 10**-21 * kg, "attogram")
fg = FG = unit("fg", 10**-18 * kg, "femtogram")
pg = PG = unit("pg", 10**-15 * kg, "picogram")
ng = NG = unit("ng", 10**-12 * kg, "nanogram")
ug = UG = unit("ug", 10**-9 * kg, "microgram")
mg = MG = unit("mg", 10**-6 * kg, "milligram")
cg = CG = unit("cg", 10**-5 * kg, "centigram")
dg = DG = unit("dg", 10**-4 * kg, "decigram")
g = unit("g", 10**-3 * kg, "gram")


# cleaning
del Unum
del unit



import wpilib


class SingleSolenoidPiston:
    """
    Wrapper class for a single solenoid piston.
    """

    def __init__(self, module, channel, ctre_hub=False):
        """
        Args:
            module (int): The Pneumatics Control Module (PCM) CAN ID
            channel (int): The channel on the PCM that the solenoid is connected to
            ctre_hub (bool): Whether the PCM is connected to a CTRE Pneumatics Control Module or not.
        """
        module_type = wpilib.PneumaticsModuleType.CTREPCM if ctre_hub else wpilib.PneumaticsModuleType.REVPH
        self.solenoid = wpilib.Solenoid(module, module_type, channel)

    def extend(self):
        """
        Extend the piston.
        """
        self.solenoid.set(True)

    def retract(self):
        """
        Retract the piston.
        """
        self.solenoid.set(False)

    def toggle(self):
        """
        Toggle the piston.
        """
        self.solenoid.toggle()

    def get_value(self) -> bool:
        """
        Get the piston state (True for extended, False for retracted).
        """
        return self.solenoid.get()



import inspect
import logging.config
import os


from robotpy_toolkit_7407.utils.color import Color, NoColor

"""
Logger utility for debugging. 

Example usage:
    utils.logger.Logger.log_info("testing")
"""


def get_default_logging():
    return {
        "version": 1,
        "formatters": {
            "standard": {
                "format": Color.RED
                + "%(asctime)s,%(msecs)d"
                + Color.END
                + Color.PURPLE
                + " %(levelname)-8s"
                + Color.END
                + " %(message)s",
                "datefmt": "%Y-%m-%d:%H:%M:%S",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "level": "INFO",
                "stream": "ext://sys.stdout",
            },
            "default": {
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
            },
            "access": {
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            },
            #         "asgi": {
            #             "formatter": "standard",
            #             "class": "logging.StreamHandler",
            #             "stream": "ext://sys.stdout",
            #         },
        },
        "loggers": {
            __name__: {"level": "INFO", "handlers": ["console"], "propagate": False,},
            "": {"handlers": ["default"], "level": "INFO"},
            "uvicorn.error": {"handlers": ["default"], "level": "INFO", "propagate": False},
            "uvicorn.access": {"handlers": ["access"], "level": "INFO", "propagate": False},
            #         "uvicorn.asgi": {"handlers": ["asgi"], "level": "TRACE", "propagate": False},
        },
    }


class Logger:
    def __init__(self, logging_config=None):
        if logging_config is None:
            logging_config = get_default_logging()

        logging.config.dictConfig(logging_config)

        self.MAX_FILENAME_LENGTH = 0

        self.root_folder = os.path.dirname(
            os.path.dirname(inspect.currentframe().f_code.co_filename)
        )

    def _log_function(
        self, func, msg: str, header=None, frame=None, traceback_length=5
    ) -> str:
        """
        Internal colored logging function.
        """
        if not frame:
            frame = inspect.currentframe().f_back

        file_name = os.path.basename(frame.f_code.co_filename)
        line_no = str(frame.f_lineno)

        caller = ""

        for i in range(traceback_length):
            if frame is None:
                break

            temp_folder_name = os.path.dirname(frame.f_code.co_filename)

            if self.root_folder == temp_folder_name:
                caller = "(" + frame.f_code.co_name + ") " + caller

                frame = frame.f_back
            else:
                continue

        if caller:
            msg = Color.GREEN + caller + Color.END + msg

        if header:
            msg = Color.YELLOW + header + Color.END + " " + msg

        filename_display = " [" + file_name + ":" + line_no + "] "
        if len(filename_display) > self.MAX_FILENAME_LENGTH:
            self.MAX_FILENAME_LENGTH = len(filename_display)

        msg = (
            Color.CYAN
            + filename_display.ljust(self.MAX_FILENAME_LENGTH)
            + Color.END
            + msg
        )

        func(msg)
        return msg

    @classmethod
    def log_info(
        cls, msg: str, header=None, frame=None, traceback_length: int = 5
    ) -> str:
        """
        Logs info
        """
        log = logging.getLogger(__name__)

        if frame:
            frame = frame.f_back
        else:
            frame = inspect.currentframe().f_back

        return cls()._log_function(log.info, str(msg), header, frame, traceback_length)

    @classmethod
    def log_error(
        cls, msg: str, header=None, frame=None, traceback_length: int = 5
    ) -> str:
        """
        Logs errors
        """
        log = logging.getLogger(__name__)
        if frame:
            frame = frame.f_back
        else:
            frame = inspect.currentframe().f_back

        return cls()._log_function(log.error, str(msg), header, frame, traceback_length)

    @classmethod
    def log_warning(
        cls, msg: str, header=None, frame=None, traceback_length: int = 5
    ) -> str:
        """
        Logs errors
        """
        log = logging.getLogger(__name__)
        if frame:
            frame = frame.f_back
        else:
            frame = inspect.currentframe().f_back

        return cls()._log_function(log.warning, str(msg), header, frame, traceback_length)

    @classmethod
    def print_function_call(cls, params=None, header="") -> str:
        """
        Prints function calls and details associated with the call
        """
        frame = inspect.currentframe().f_back

        if params:
            return cls().log_info(
                "Called "
                + inspect.getmodule(frame).__name__
                + "."
                + frame.f_code.co_name
                + " with parameters: "
                + str(params),
                header,
                frame,
            )
        else:
            return cls().log_info(
                "Called "
                + inspect.getmodule(frame).__name__
                + "."
                + frame.f_code.co_name,
                header,
                frame,
            )


info = Logger.log_info
error = Logger.log_error
warning = Logger.log_warning
warn = warning



from dataclasses import dataclass

import wpilib

DRIVER_CONTROLLER_ID = 0
OPERATOR_CONTROLLER_ID = 1


class _Joysticks:
    joysticks: list[wpilib.Joystick] = [
        wpilib.Joystick(DRIVER_CONTROLLER_ID),
        wpilib.Joystick(OPERATOR_CONTROLLER_ID)
    ]


@dataclass
class JoystickAxis:
    """
    Wrapper for wpilib joystick button
    """
    controller_id: int
    axis_id: int

    @property
    def value(self) -> float:
        """
        Gets the value of the axis selected

        Returns:
            float: The value between -1 and 1 as a float
        """
        return _Joysticks.joysticks[self.controller_id].getRawAxis(self.axis_id)



from robotpy_toolkit_7407.unum import Unum, units

# --- UNUMS ---
m = units.m
ft = Unum.unit("ft", 0.3048 * m, "foot")
inch = Unum.unit("in", ft / 12, "inch")
mile = units.mile

s = units.s
ms = units.ms
minute = units.min
hour = units.h

rad = units.rad
deg = units.deg
rev = Unum.unit("rev", 360 * deg, "revolution")


# --- TYPING ---
meters = float
radians = float
seconds = float
meters_per_second = float
meters_per_second_squared = float
radians_per_second = float
radians_per_second_squared = float
radians_per_meter = float
meters_per_radian = float
rotations_per_second = float
rotations = float


import math
from wpimath.geometry import Pose3d, Rotation3d, Translation3d

dist_scalar = 1

field_to_target = Pose3d(Translation3d(2, -2, 1), Rotation3d(0, 0, 0))
camera_to_target = Pose3d(Translation3d(1, -1, 0), Rotation3d(0, 0, math.radians(45)))
roc = (.5, -.5)
robot_to_camera = Pose3d(Translation3d(roc[0], roc[1], 1), Rotation3d(0, 0, math.atan2(roc[0], roc[1])))

gyro_angle = math.radians(45)

field_to_camera = (
    field_to_target.x - camera_to_target.x,
    field_to_target.y - camera_to_target.y
)

print("FIELD TO CAMERA: ", field_to_camera)

theta = robot_to_camera.rotation().angle - gyro_angle

print("THETA: ", math.degrees(theta))

h = (robot_to_camera.x ** 2 + robot_to_camera.y ** 2) ** .5

field_to_robot = (
    field_to_camera[0] - h * math.sin(theta),
    field_to_camera[1] - h * math.cos(theta)
)

print("FIELD TO ROBOT: ", field_to_robot)



import photonvision
from robotpy_apriltag import AprilTag
from wpimath.geometry import Pose3d


class PhotonTarget:
    def __init__(self, target: photonvision.PhotonTrackedTarget):
        self.raw_target = target
        self.relative_pose = self.raw_target.getBestCameraToTarget()
        self.ID = self.raw_target.getFiducialId()
        self.pitch = self.raw_target.getPitch()
        self.yaw = self.raw_target.getYaw()


class AprilTag(AprilTag):
    def __init__(self, ID: int, pose: Pose3d):
        super().__init__()
        self.ID = ID
        self.pose = pose



from robotpy_toolkit_7407.motors import ctre_motors
from robotpy_toolkit_7407.unum import Unum

from robotpy_toolkit_7407.command import SubsystemCommand, T
from robotpy_toolkit_7407.motors.ctre_motors import talon_sensor_vel_unit
from robotpy_toolkit_7407.subsystem_templates.drivetrain.differential_drivetrain import DifferentialDrivetrain
from robotpy_toolkit_7407.utils.math import clamp, talon_sensor_units_to_inches, inches_to_talon_sensor_units
from robotpy_toolkit_7407.utils.units import m, s


# TODO Redo this to make it like swerve drivetrain commands
class DriveArcade(SubsystemCommand[DifferentialDrivetrain]):
    """
    Arcade drive command for differential drivetrain.
    """
    def __init__(self, subsystem: T, track_width_inches: float):
        """

        Args:
            subsystem: drivetrain
            track_width_inches: track width of the robot in inches
        """
        super().__init__(subsystem)
        self.track_width_inches = track_width_inches

    def initialize(self) -> None:
        pass

    def execute(self) -> None:
        x_axis, y_axis = self.subsystem.axis_x.value, self.subsystem.axis_y.value

        x_axis, y_axis = self._add_dead_zones(x_axis, y_axis)

        left, right = self._turn_radius_drive(x_axis, y_axis, self.track_width_inches)

        self.subsystem.set_motor_velocity(left * ctre_motors.k_sensor_vel_to_rad_per_sec, -right * ctre_motors.k_sensor_vel_to_rad_per_sec)

    def end(self, interrupted: bool) -> None:
        self.subsystem.set_motor_velocity(0, 0)

    def isFinished(self) -> bool:
        return False

    def runsWhenDisabled(self) -> bool:
        return False

    @staticmethod
    def _add_dead_zones(x_axis: float, y_axis: float) -> tuple[float, float]:
        if abs(x_axis) < 0.2:
            x_axis = 0
        if abs(y_axis) < 0.2:
            y_axis = 0

        return x_axis, y_axis

    @staticmethod
    def _arcade_drive(x_axis: float, y_axis: float) -> tuple[float, float]:
        left = clamp(y_axis + x_axis, -1, 1)
        right = clamp(y_axis - x_axis, -1, 1)

        # left *= 18000
        # right *= 18000

        return left, right

    @staticmethod
    def _turn_radius_drive(x_axis: float, y_axis: float, track_width_inches: float) -> tuple[float, float]:
        if y_axis > 0:
            x_axis = -x_axis

        # x_axis *= abs(x_axis)

        velocity_sensor_units = 18000 * y_axis
        target_velocity = talon_sensor_units_to_inches(velocity_sensor_units, True)

        if x_axis > 0:
            turn_radius = 120.0 * (1.05 - x_axis)
        elif x_axis < 0:
            turn_radius = 120.0 * (-1.05 - x_axis)
        else:
            turn_radius = 0

        if x_axis == 0:
            velocity_difference = 0
        elif target_velocity == 0 or turn_radius == 0:
            velocity_difference = talon_sensor_units_to_inches(-18000.0 * x_axis, True)
        else:
            velocity_difference = (track_width_inches * target_velocity) / turn_radius

        left = inches_to_talon_sensor_units(target_velocity - velocity_difference, True)
        right = inches_to_talon_sensor_units(target_velocity + velocity_difference, True)

        left = clamp(left, -18000, 18000)
        right = clamp(right, -18000, 18000)

        return left, right



from robotpy_toolkit_7407.utils.units import radians


class Gyro:
    """
    Extendable class for swerve gyro.
    """

    def init(self, gyro_start_angle=0):
        """
        Initialize the swerve gyro. Overridden class.
        """
        ...

    def get_robot_heading(self) -> radians:
        """
        Get the robot heading in radians. Overridden class. Must return radians.
        """
        ...

    def reset_angle(self):
        """
        Reset the robot heading. Overridden class.
        """
        ...



from dataclasses import dataclass

import commands2 as commands
import commands2.button

from robotpy_toolkit_7407.oi.joysticks import _Joysticks


@dataclass
class _Button:
    controller_id: int

    def __call__(self) -> commands.button.Button: ...


@dataclass
class DefaultButton(_Button):
    """
    Wrapper for wpilib button
    """
    button_id: int

    def __call__(self) -> commands.button.Button:
        if self.button_id < 0:
            return commands.button.Button(
                lambda: _Joysticks.joysticks[self.controller_id].getRawAxis(-self.button_id) > 0.8
            )
        return commands.button.JoystickButton(_Joysticks.joysticks[self.controller_id], self.button_id)


@dataclass
class AxisButton(_Button):
    """
    Wrapper for wpilib axis button
    """
    axis_id: int
    range_min: float = -1
    range_max: float = 1

    def __call__(self) -> commands.button.Button:
        return commands.button.Button(
            lambda: self.range_min <= _Joysticks.joysticks[self.controller_id].getRawAxis(self.axis_id) <= self.range_max
        )



from photonvision import PhotonTrackedTarget
from robotpy_apriltag import AprilTagFieldLayout
from wpimath.geometry import Pose3d, Translation3d, Rotation3d, Transform3d

from robotpy_toolkit_7407.sensors.gyro import PigeonIMUGyro_Wrapper
from robotpy_toolkit_7407.sensors.photonvision import PhotonCamera, PhotonTarget, PhotonOdometry, AprilTag

gyro = PigeonIMUGyro_Wrapper(13)

camera = PhotonCamera("globalshuttercamera",
                      Pose3d(Translation3d(x=1, y=1, z=.65),
                             Rotation3d(roll=0, pitch=0, yaw=0)),
                      scale_constant=1)

current_target = PhotonTarget(
    PhotonTrackedTarget(
        1, 1, 1, 1, 1,
        Transform3d(
            Pose3d(0, 0, 0, Rotation3d(0, 0, 0)),
            Pose3d(1, 1, 1, Rotation3d(0, 0, 0))),
        Transform3d(
            Pose3d(1, 1, 1, Rotation3d(1, 1, 1)),
            Pose3d(0, 0, 0, Rotation3d(0, 0, 0))),
        .1,
        [(1, 1), (1, 1), (1, 1), (1, 1)]))

field_layout = {
    'apriltags': {
        1: Pose3d(
            Translation3d(x=4.1783, y=1.2065, z=0.70485),
            Rotation3d(roll=0, pitch=0, yaw=0)
        ),
    },
    'fieldLength': 50,
    'fieldWidth': 30
}

odometry = PhotonOdometry(
    camera,
    field_layout,
    gyro
)

odometry.refresh()
print(odometry.getRobotPose(target=current_target))



"""Define the SI derived units.

Source: http://physics.nist.gov/cuu/Units/units.html
"""

from robotpy_toolkit_7407.unum import Unum
from robotpy_toolkit_7407.unum.units.si.base import *
unit = Unum.unit

rad        = RAD     = unit( 'rad'   , M / M          , 'radian'         )
sr         = SR      = unit( 'sr'    , M**2 / M**2    , 'steradian'      )
Hz         = HZ      = unit( 'Hz'    , 1 / s          , 'hertz'          )
N          = N       = unit( 'N'     , M*KG / s**2    , 'newton'         )
Pa         = PA      = unit( 'Pa'    , N / M**2       , 'pascal'         )
J          = J       = unit( 'J'     , N*M            , 'joule'          )
W          = W       = unit( 'W'     , J / s          , 'watt'           )
C          = C       = unit( 'C'     , s * A          , 'coulomb'        )
V          = V       = unit( 'V'     , W / A          , 'volt'           )
F          = F       = unit( 'F'     , C / V          , 'farad'          )
ohm        = OHM     = unit( 'ohm'   , V / A          , 'ohm'            )
S          = SIEMENS = unit( 'S'     , A / V          , 'siemens'        )
Wb         = WB      = unit( 'Wb'    , V * s          , 'weber'          )
T          = T       = unit( 'T'     , WB / M**2      , 'tesla'          )
H          = HENRY   = unit( 'H'     , WB / A         , 'henry'          )
# warning : conversion assumes relative temperatures
celsius    = CELSIUS = unit( 'deg C' , K              , 'degree Celsius' )  
lm         = LM      = unit( 'lm'    , CD * SR        , 'lumen'          )
lx         = LX      = unit( 'lx'    , LM / M**2      , 'lux'            )
Bq         = BQ      = unit( 'Bq'    , 1 / s          , 'becquerel'      )
Gy         = GY      = unit( 'Gy'    , J / KG         , 'gray'           )
Sv         = SV      = unit( 'Sv'    , J / KG         , 'sievert'        )
kat        = KAT     = unit( 'kat'   , MOL / s        , 'katal'          )

# cleaning
del Unum
del unit



import math
import time

from wpimath.controller import HolonomicDriveController, PIDController, ProfiledPIDControllerRadians
from wpimath.geometry import Rotation2d
from wpimath.trajectory import TrapezoidProfileRadians, Trajectory

from robotpy_toolkit_7407.command import SubsystemCommand
from robotpy_toolkit_7407.subsystem_templates.drivetrain.swerve_drivetrain import SwerveDrivetrain
from robotpy_toolkit_7407.utils import logger
from robotpy_toolkit_7407.utils.math import rotate_vector, bounded_angle_diff
from robotpy_toolkit_7407.utils.units import m, s, rad


class DriveSwerve(SubsystemCommand[SwerveDrivetrain]):
    """
    Drive the robot using a swerve drive controller.
    """

    def initialize(self) -> None:
        pass

    def execute(self) -> None:
        """
        Execute the command. Can be overridden for a custom swerve drive.
        """
        dx, dy, d_theta = self.subsystem.axis_dx.value, self.subsystem.axis_dy.value, self.subsystem.axis_rotation.value

        dx *= self.subsystem.max_vel.asUnit(m / s)
        dy *= -self.subsystem.max_vel.asUnit(m / s)

        self.subsystem.set_driver_centric((dx, dy), -d_theta * self.subsystem.max_angular_vel)

    def end(self, interrupted: bool) -> None:
        self.subsystem.stop()

    def isFinished(self) -> bool:
        return False

    def runsWhenDisabled(self) -> bool:
        return False


class FollowPath(SubsystemCommand[SwerveDrivetrain]):
    """
    Follow a given wpimath trajectory using a swerve drive controller.
    """
    def __init__(self, subsystem: SwerveDrivetrain, trajectory: Trajectory, period: float = 0.02):
        super().__init__(subsystem)
        self.trajectory = trajectory
        self.controller = HolonomicDriveController(
            PIDController(1, 0, 0),
            PIDController(1, 0, 0),
            ProfiledPIDControllerRadians(
                8, 0, 0, TrapezoidProfileRadians.Constraints(
                    subsystem.max_angular_vel,
                    subsystem.max_angular_vel / .01
                ), period
            )
        )
        self.start_time = 0
        self.t = 0
        self.duration = trajectory.totalTime()
        self.theta_i = trajectory.initialPose().rotation().radians()
        self.theta_f = trajectory.sample(self.duration).pose.rotation().radians()
        self.theta_diff = bounded_angle_diff(self.theta_i, self.theta_f)
        self.omega = self.theta_diff / self.duration

    def initialize(self) -> None:
        self.start_time = time.perf_counter()

    def execute(self) -> None:
        self.t = time.perf_counter() - self.start_time
        if self.t > self.duration:
            self.t = self.duration
        goal = self.trajectory.sample(self.t)
        goal_theta = self.theta_i + self.omega * self.t
        speeds = self.controller.calculate(self.subsystem.odometry.getPose(), goal, Rotation2d(goal_theta))
        vx, vy = rotate_vector(
            speeds.vx, speeds.vy,
            self.subsystem.odometry.getPose().rotation().radians()
        )
        self.subsystem.set_driver_centric((vx, vy), speeds.omega)

    def end(self, interrupted: bool) -> None:
        pass

    def isFinished(self) -> bool:
        return self.t > self.trajectory.totalTime()

    def runsWhenDisabled(self) -> bool:
        return False



class XBoxController:
    """
    Mappings for XBox Controller ports
    """
    A = 1
    B = 2
    X = 3
    Y = 4
    LB = 5
    RB = 6
    LT = -2
    RT = -3
    SELECT = 7
    START = 8
    L_JOY = (0, 1)  # (x, y)
    R_JOY = (4, 5)  # (x, y)


class LogitechController:
    """
    Mappings for Logitech Controller ports
    """
    A = 2
    B = 3
    X = 1
    Y = 4
    LB = 5
    RB = 6
    LT = 7
    RT = 8
    SELECT = 9
    START = 10
    L_JOY = (0, 1)  # (x, y)
    R_JOY = (2, 3)  # (x, y)



"""Define common units that are not SI units.

Source: http://physics.nist.gov/cuu/Units/outside.html
"""
from math import pi
from robotpy_toolkit_7407.unum import Unum
from robotpy_toolkit_7407.unum.units.si import *
unit = Unum.unit

min        = MIN      = unit( 'min'      , 60 * s            , 'minute'                    )
h          = H        = unit( 'h'        , 60 * MIN          , 'hour'                      )
d          = D        = unit( 'd'        , 24 * H            , 'day'                       )
deg        = ARCDEG   = unit( 'deg'      , pi/180 * RAD      , 'degree (angle)'            )
arcmin     = ARCMIN   = unit( "'"        , ARCDEG / 60       , 'minute (angle)'            )
arcsec     = ARCSEC   = unit( "''"       , ARCMIN / 60       , 'second (angle)'            )
l          = L        = unit( 'L'        , 1E-3 * M**3       , 'liter'                     )
t          = TON      = unit( 't'        , 1E3 * KG          , 'metric ton'                )
Np         = NP       = unit( 'Np'       , 1                 , 'neper'                     )
dB         = DECIBEL  = unit( 'dB'       , 0                 , 'decibel'                   )
eV         = EV       = unit( 'eV'       , 1.60218E-19 * J   , 'electronvolt'              )
u          = U        = unit( 'u'        , 1.66054E-27 * KG  , 'unified atomic mass unit'  )
ua = AU    = UA       = unit( 'ua'       , 1.49598E11 * M    , 'astronomical unit'         )
mile       = MILE     = unit( 'mile'     , 1609.34 * M       , 'statute mile'              )
nmile      = NMILE    = unit( 'nmi'      , 1852 * M          , 'nautical mile'             )
knot       = KNOT     = unit( 'knot'     , MILE / H          , 'knot'                      )
a          = ARE      = unit( 'a'        , 1E2 * M**2        , 'are'                       )
ha         = HA       = unit( 'ha'       , 1E4 * M**2        , 'hectare'                   )
bar        = BAR      = unit( 'bar'      , 1E5 * PA          , 'bar'                       )
angstrom   = ANGSTROM = unit( 'angstrom' , 1E-10 * M         , 'angstrom'                  )
b          = B        = unit( 'b'        , 1E-28 * M**2      , 'barn'                      )
Ci         = CI       = unit( 'Ci'       , 3.7E10 * BQ       , 'curie'                     )
R          = R        = unit( 'R'        , 2.58E-4 * C / KG  , 'roentgen'                  )
rem        = REM      = unit( 'rem'      , 1E-2 * SV         , 'rem'                       )

# Note : 'rad' defined as 1E-2 Gy as been left out because it conflits with
# using 'rad' for radians.

# cleaning
del Unum
del unit
del pi